//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  node.hpp
//! \brief nodes within expression trees, inheriting from virtual base CNode

#ifndef NODE_H
#define NODE_H

//! Base class CNode is declared elsewhere (expr.hpp).


//! Node with a few arguments. Pure virtual base class.

class CNodeWithArgs : public CNode
{
protected:
    int narg;
    vector<RNode> arg;

public:
    CNodeWithArgs(int _narg) : narg(_narg) { arg.resize(narg); }
    int npar() const;
    bool k_dependent() const;
    bool j_dependent() const;
    bool i_dependent() const;
    bool has_dummy() const;
    bool has_conv() const;
    virtual string tree_info() const = 0;
};


//! Generic function/operator node. Pure virtual base class.

class CNodeFun : public CNodeWithArgs
{
protected:
    const class CFunc* fu;

public:
    CNodeFun(int _narg, const class CFunc* _fu) : CNodeWithArgs(_narg), fu(_fu) {}
    RObj tree_val(const CContext& ctx) const;
    virtual CCoord node_coord(int k) const = 0;
    virtual string tree_info() const = 0;
};


//! Function/operator node with one argument.

class CNodeFun1 : public CNodeFun
{
public:
    CNodeFun1(const class CFunc* _fu, RNode a0);
    CCoord node_coord(int k) const;
    string tree_info() const;
};


//! Function/operator node with two arguments.

class CNodeFun2 : public CNodeFun
{
public:
    CNodeFun2(const class CFunc* _fu, RNode a0, RNode a1);
    CCoord node_coord(int k) const;
    string tree_info() const;
};


//! Function/operator node with three arguments.

class CNodeFun3 : public CNodeFun
{
public:
    CNodeFun3(const class CFunc* _fu, RNode a0, RNode a1, RNode a2);
    CCoord node_coord(int k) const;
    string tree_info() const;
};


//! Range node.

class CNodeRange : public CNodeWithArgs
{
private:
    bool inclusive_end;

public:
    CNodeRange(
        bool _inclusive_end, RNode _beg = nullptr, RNode _end = nullptr, RNode _step = nullptr)
        : CNodeWithArgs(3), inclusive_end(_inclusive_end)
    {
        arg[0] = _beg;
        arg[1] = _end;
        arg[2] = _step;
    }
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const;
    bool looks_like_indices() const { return !arg[0] || arg[0]->looks_like_indices(); }
};


//! List node.

class CNodeList : public CNodeWithArgs
{
public:
    CNodeList(RNode _list, RNode _incr);
    vector<RNode> args() const { return arg; }
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const;
    bool looks_like_indices() const { return arg[0] && arg[0]->looks_like_indices(); }
};


//! Generalized integral node.

class CNodeGeni : public CNodeWithArgs
{
private:
    const class CGeni* geni;

public:
    CNodeGeni(const class CGeni* _geni, RRef _ref = nullptr);
    CNodeGeni(const class CGeni* _geni, RNode a0);
    CNodeGeni(const class CGeni* _geni, RNode a0, RNode a1);
    bool k_dependent() const { return true; }
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const;
};


//! Curve functional node.

class CNodeCvin : public CNodeWithArgs
{
private:
    const class CCvin* cvin;
    RRef ref;

public:
    CNodeCvin(const class CCvin* _cvin, RRef _ref, RNode a0, RNode a1);
    bool k_dependent() const { return true; }
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const;
};


//! End node with fixed numeric value.

class CNodeVal : public CNode
{
private:
    RObj val;

public:
    CNodeVal(RObj _val) : val(_val) {}
    CNodeVal(double _val);
    CNodeVal(int _val);
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const;
    bool looks_like_indices() const;
};


//! Index node base class.

class CNodeIdx : public CNode
{
public:
    bool k_dependent() const { return true; }
};


//! Index i node.

class CNodeIdxI : public CNodeIdx
{
public:
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const { return CCoord("i", ""); }
    string tree_info() const { return "i"; }
    bool i_dependent() const { return true; }
};


//! Index j node.

class CNodeIdxJ : public CNodeIdx
{
public:
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const { return CCoord("j", ""); }
    string tree_info() const { return "j"; }
    bool j_dependent() const { return true; }
};


//! Index k node.

class CNodeIdxK : public CNodeIdx
{
public:
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const { return CCoord("k", ""); }
    string tree_info() const { return "k"; }
};


//! Session-wide workspace information: NK = number of online files.

class CNodeSessionNK : public CNode
{
public:
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const { return CCoord("nk", ""); }
    string tree_info() const { return "nk"; }
};


//! File-wide information.

class CNodeFile : public CNode
{
private:
    virtual RObj tree_val_scalar(const CContext& ctx) const
    {
        throw S("BUG: unforeseen call to CNodeFile::tree_val_scalar");
    }
    virtual CCoord file_coord(const COlo* f) const { return CCoord(name(), ""); }
protected:
    RRef ref;

public:
    CNodeFile(RRef _ref) : ref(_ref) {}
    void set_ref(RRef& _ref) { ref = _ref; };
    bool k_dependent() const { return true; }
    RObj tree_val(const CContext& ctx) const;
    string tree_info() const { return name() + ref->ref_info(); }
    CCoord node_coord(int k) const;
    virtual string name() const = 0;
};


//! File-wide information: NJ = number of slices.

class CNodeFileNJ : public CNodeFile
{
public:
    CNodeFileNJ(RRef _ref) : CNodeFile(_ref) {}
    string name() const { return "nj"; }
    RObj tree_val_scalar(const CContext& ctx) const;
};


//! File-wide information: R# = real parameter.

class CNodeFileR : public CNodeFile
{
private:
    int num;

public:
    CNodeFileR(int _num, RRef _ref) : CNodeFile(_ref), num(_num) {}
    CCoord file_coord(const COlo* f) const;
    string name() const { return "r" + S(num); }
    RObj tree_val_scalar(const CContext& ctx) const;
};


//! Slice-wide information.

class CNodeSlice : public CNodeFile
{
public:
    CNodeSlice(RRef _ref) : CNodeFile(_ref) {}
};


//! Slice-wide information: Z# = z coordinate.

class CNodeSliceZ : public CNodeSlice
{
private:
    int num;

public:
    CNodeSliceZ(int _num, RRef _ref) : CNodeSlice(_ref), num(_num) {}
    CCoord file_coord(const COlo* f) const;
    string name() const { return "z" + S(num); }
    RObj tree_val_scalar(const CContext& ctx) const;
};


//! Spec-wide information: number of data points.

class CNodeSpecNI : public CNodeSlice
{
public:
    CNodeSpecNI(RRef _ref) : CNodeSlice(_ref) {}
    string name() const { return "ni"; }
    RObj tree_val_scalar(const CContext& ctx) const;
};


//! Curve-wide information.

class CNodeCurve : public CNodeSlice
{
public:
    CNodeCurve(RRef _ref) : CNodeSlice(_ref) {}
    RObj tree_val_scalar(const CContext& ctx) const;
    virtual RObj curve_val_scalar(const CCurve* c) const = 0;
};


//! Curve-wide information: P# = curve parameter.

class CNodeCurveP : public CNodeCurve
{
private:
    int num;

public:
    CNodeCurveP(int _num, RRef _ref) : CNodeCurve(_ref), num(_num) {}
    CCoord file_coord(const COlo* f) const;
    string name() const { return "p" + S(num); }
    RObj curve_val_scalar(const CCurve* c) const;
    int npar() const { return num + 1; }
};


//! Curve-wide information: fit outcome code.

class CNodeCurveOutcome : public CNodeCurve
{
public:
    CNodeCurveOutcome(RRef _ref) : CNodeCurve(_ref) {}
    string name() const { return "oc"; }
    RObj curve_val_scalar(const CCurve* c) const;
};


//! Curve-wide information: fit outcome chi^2.

class CNodeCurveChi2 : public CNodeCurve
{
public:
    CNodeCurveChi2(RRef _ref) : CNodeCurve(_ref) {}
    string name() const { return "chi2"; }
    RObj curve_val_scalar(const CCurve* c) const;
};


//! Curve-wide information: fit outcome R^2.

class CNodeCurveR2 : public CNodeCurve
{
public:
    CNodeCurveR2(RRef _ref) : CNodeCurve(_ref) {}
    string name() const { return "R2"; }
    RObj curve_val_scalar(const CCurve* c) const;
};


//! Pointwise information.

class CNodePoint : public CNodeSlice
{
private:
    RObj tree_val_point(const CContext& ctx, int i) const;
    RObj tree_val_vector(const CContext& ctx) const;

protected:
    virtual RObj spec_val_point(const CSpec* s, int i, bool want_error) const = 0;
    virtual RObj spec_val_vector(const CSpec* s, int n, bool want_error) const = 0;

public:
    CNodePoint(RRef _ref) : CNodeSlice(_ref) {}
    RObj tree_val(const CContext& ctx) const;
};


//! Pointwise information: X.

class CNodePointX : public CNodePoint
{
public:
    CNodePointX(RRef _ref) : CNodePoint(_ref) {}
    CCoord file_coord(const COlo* f) const;
    string name() const { return "x"; }
    RObj spec_val_point(const CSpec* s, int i, bool want_error) const;
    RObj spec_val_vector(const CSpec* s, int n, bool want_error) const;
};


//! Pointwise information: Y.

class CNodePointY : public CNodePoint
{
public:
    CNodePointY(RRef _ref) : CNodePoint(_ref) {}
    CCoord file_coord(const COlo* f) const;
    string name() const { return "y"; }
    RObj spec_val_point(const CSpec* s, int i, bool want_error) const;
    RObj spec_val_vector(const CSpec* s, int n, bool want_error) const;
};


//! Pointwise information: DY.

class CNodePointDY : public CNodePoint
{
public:
    CNodePointDY(RRef _ref) : CNodePoint(_ref) {}
    CCoord file_coord(const COlo* f) const;
    string name() const { return "dy"; }
    RObj spec_val_point(const CSpec* s, int i, bool want_error) const;
    RObj spec_val_vector(const CSpec* s, int n, bool want_error) const;
};


//! Identifier node.

class CNodeIdf : public CNode
{
private:
    string idf;

public:
    CNodeIdf(string _s);
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const { return "$"; }
};


//! Function argument (t) node.

class CNodeDummy : public CNode
{
public:
    CNodeDummy() {}
    RObj tree_val(const CContext& ctx) const;
    bool has_dummy() const { return true; } // this node _is_ t, so it has t
    CCoord node_coord(int k) const
    {
        throw "BUG: CNodeDummy::node_coord not expected to be needed";
    }
    string tree_info() const { return "farg"; }
};


//! Curve evaluation node.

class CNodeCev : public CNode
{
private:
    RRef ref;
    RNode arg;

public:
    CNodeCev(RRef _ref, RNode _arg) : ref(_ref), arg(_arg) {}
    bool k_dependent() const { return true; }
    bool has_dummy() const { return arg->has_dummy(); }
    bool has_conv() const { return arg->has_conv(); }
    RObj tree_val(const CContext& ctx) const;
    CCoord node_coord(int k) const;
    string tree_info() const { return "c[" + ref->ref_info() + "]"; }
};


//! Base class for mixing in data from a "convolution" file

class CNodeMixin : public CNode
{
protected:
    RNode shift;

public:
    CNodeMixin(const RNode& _shift) : shift(_shift) {}
    int npar() const { return shift->npar(); }
    bool k_dependent() const { return shift->k_dependent(); }
    bool j_dependent() const { return shift->j_dependent(); }
    bool i_dependent() const { return shift->i_dependent(); }
    RObj tree_val(const CContext& ctx) const;
    virtual RObj copy_theory(const CContext& ctx, double theshift) const = 0;
    virtual RObjVecDbl convolve(
        const CContext& ctx, double theshift, const CSpec* sv, double conv_norm,
        double conv_step) const = 0;
    bool has_conv() const { return true; }
};

//! Base class for convoluting a function with a spectrum.

class CNodeConvBase : public CNodeMixin
{
protected:
    RNode theory;

public:
    CNodeConvBase(const RNode& _theory, const RNode& _shift) : CNodeMixin(_shift), theory(_theory)
    {
    }
    int npar() const { return std::max(shift->npar(), theory->npar()); }
    bool k_dependent() const { return theory->k_dependent() || shift->k_dependent(); }
    bool j_dependent() const { return theory->j_dependent() || shift->j_dependent(); }
    bool i_dependent() const { return theory->i_dependent() || shift->i_dependent(); }
    bool has_dummy() const { return theory->has_dummy(); }
    CCoord node_coord(int k) const { return theory->node_coord(k); }
};


//! Convolute a function with a spectrum.

class CNodeConv : public CNodeConvBase
{
public:
    CNodeConv(const RNode& _theory, const RNode& _shift = RNode(new CNodeVal(0.0)))
        : CNodeConvBase(_theory, _shift)
    {
    }
    RObj copy_theory(const CContext& ctx, double theshift) const;
    RObjVecDbl convolve(
        const CContext& ctx, double theshift, const CSpec* sv, double conv_norm,
        double conv_step) const;
    string tree_info() const
    {
        return "conv(" + theory->tree_info() + "," + shift->tree_info() + ")";
    }
};


//! Convolution a function with a spectrum, using the function's primitive.

class CNodePConv : public CNodeConvBase
{
public:
    CNodePConv(const RNode& _theory, const RNode& _shift = RNode(new CNodeVal(0.0)))
        : CNodeConvBase(_theory, _shift){};
    RObj copy_theory(const CContext& ctx, double theshift) const;
    RObjVecDbl convolve(
        const CContext& ctx, double theshift, const CSpec* sv, double conv_norm,
        double conv_step) const;
    string tree_info() const
    {
        return "pconv(" + theory->tree_info() + "," + shift->tree_info() + ")";
    }
};


//! Dirac's delta function node, to request a copy of the resolution function.

class CNodeDirac : public CNodeMixin
{
public:
    CNodeDirac(const RNode& _shift = RNode(new CNodeVal(0.0))) : CNodeMixin(_shift){};
    RObj copy_theory(const CContext& ctx, double theshift) const;
    RObjVecDbl convolve(
        const CContext& ctx, double theshift, const CSpec* sv, double conv_norm,
        double conv_step) const;
    bool has_dummy() const { return true; } // has implicit t dependence
    CCoord node_coord(int k) const;
    string tree_info() const { return "resol(" + shift->tree_info() + ")"; }
};

#endif // NODE_H
